/**
  Raytracing demo application.
  
  Orion Sky Lawlor, olawlor@acm.org, 2010-01-22 (Public Domain)
*/
#include <iostream>
#include <cmath>
#include <GL/glew.h> /* 
OpenGL Extensions Wrangler:  for gl...ARB extentions.  Must call glewInit after glutCreateWindow! */
#include <GL/glut.h> /* OpenGL Utilities Toolkit, for GUI tools */
#include "ogl/glsl.h"
#include "ogl/glsl.cpp"
#include "ogl/minicam.h"

void display(void) 
{
	glClearColor(0.4,0.5,0.7,0.0); // background color
	glClear(GL_COLOR_BUFFER_BIT + GL_DEPTH_BUFFER_BIT);
	
	glEnable(GL_DEPTH_TEST);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_BLEND);
	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity(); // flush any ancient matrices
	gluPerspective(70.0, // fov
		glutGet(GLUT_WINDOW_WIDTH)/(float)glutGet(GLUT_WINDOW_HEIGHT),
		0.1,
		1000.0); // z clipping planes
	oglCameraLookAt(1.0);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity(); // flush any ancient matrices
	
	
	/* Grid on the ground */
	double res=10.0;
	glEnable(GL_LINE_SMOOTH);
	glLineWidth(0.1);
	glBegin(GL_LINES);
	glColor3f(0,0,0);
	for (float t=-res;t<=res;t+=1.0) {
		float z=0.0;
		glVertex3f(t,-res,z); // -Y
		glVertex3f(t,+res,z); // +Y
		glVertex3f(-res,t,z); // -X
		glVertex3f(+res,t,z); // +X
		
	}
	glEnd();
	
	
	/* The proxy sphere should be gently rotating */
	const float rotate_rate=0.1; /* revolutions per second */
	float angle=glutGet(GLUT_ELAPSED_TIME)*0.001*rotate_rate*360;
	glRotatef(angle,0,0,1);
	
	/* Build a pixel shader */
	static GLhandleARB prog=MAKE_PROGRAM_OBJECT(
		// GLSL vertex shader!
		varying vec3 G; // proxy geometry location, world coordinates
		void main(void) {
			vec4 mv=gl_ModelViewMatrix*gl_Vertex; // apply modelview to get to world coords
			G=vec3(mv);
			gl_Position = gl_ProjectionMatrix*mv;
		}
	,
		// GLSL fragment shader
		uniform vec3 C; // camera position, world coordinates
		varying vec3 G; // proxy geometry location, world coordinates
		uniform float k;
		void main(void) {
			vec3 D=normalize(G-C); // points from camera to proxy
			
			// Rendering a "notaboloid", z^2 + k = x^2 + y^2
			
			// Quadratic equation terms:
			float a=D.x*D.x+D.y*D.y-D.z*D.z; // t^2 term
			float b=2.0*C.x*D.x+2.0*C.y*D.y-2.0*C.z*D.z;  // t term
			float c=C.x*C.x+C.y*C.y-C.z*C.z-k; // constant term

			// Solve the quadratic equation:
			float det=b*b-4.0*a*c;
			if (det<0.0) discard; /* miss */
			float t=(-b-sqrt(det))/(2.0*a); /* FIXME: + branch too? */
			if (t<0.0) discard; /* behind your head */
			vec3 P=C+t*D;
			if (abs(P.z)>1.0) discard; /* out of bounds */
			
			// Light this hit point
			vec3 gradient=vec3(2.0*P.x,2.0*P.y,-2.0*P.z);
			vec3 N = normalize(gradient); // Points away from surface
			vec3 L1 = normalize(vec3(-0.8,0.0,0.2)); // Points toward light 1
			vec3 T = normalize(vec3(C)-P); // Points from intersection Toward camera
			vec3 H1 = normalize(T+L1); // Blinn's 'halfway vector' for light 1
			float amb=0.1; // Ambient illumination
			float d=clamp(dot(N,L1),0.0,1.0); // Diffuse light fraction
			float s=pow(clamp(dot(N,H1),0.0,1.0),50.0); // Specular fraction.  Constant controls highlight size
			
			vec4 M=vec4(0.9,0.8,0.2,1.0); /* material color: yellow */
			vec3 checker=step(0.5,fract(P*10.0))-0.5; /* checkerboard */
			vec4 K=vec4(step(checker.x*checker.y*checker.z,0.0));
			
			gl_FragColor = M*K*(d+amb)+vec4(s);
			gl_FragColor.a=1.0; /* always be opaque */
		}
	);
	glUseProgramObjectARB(prog);
	glFastUniform3fv(prog,"C",1,camera);
	glFastUniform1f(prog,"k",0.2*sin(angle/180.0*3.0));
	
	glutSolidSphere(2.0,8,6); // proxy geometry
	glUseProgramObjectARB(0);
	
	glutPostRedisplay(); // continual animation
	glutSwapBuffers();
}

int main(int argc,char *argv[]) 
{
	glutInit(&argc,argv);
	glutInitDisplayMode(GLUT_RGBA + GLUT_DEPTH + GLUT_DOUBLE);
	glutInitWindowSize(800,600);
	glutCreateWindow("Raytrace + Lighting");
	
	glewInit();
	
	glutDisplayFunc(display);
	oglCameraInit();
	camera=vec3(0,-3,1);
	
	glutMainLoop();
	return 0;
}
